
-- CotW AI functions --

local M = {}

local function estDistCost(avgCost, x1, y1, x2, y2)
	return (math.abs(x1-x2) + math.abs(y1-y2)) * avgCost
end

local function getNeighbors(map, x, y) -- *** Limited to current map only -- change later?
	local neighbors = {}
	if map[x-1] and map[x-1][y] then neighbors[#neighbors+1] = {x=x-1, y=y} end
	if map[x+1] and map[x+1][y] then neighbors[#neighbors+1] = {x=x+1, y=y} end
	if map[x][y-1] then neighbors[#neighbors+1] = {x=x, y=y-1} end
	if map[x][y+1] then neighbors[#neighbors+1] = {x=x, y=y+1} end
	if map[x][y].door and map[x][y].door.map == map then neighbors[#neighbors+1] = {x=map[x][y].door.x, y=map[x][y].door.y} end -- FIX THIS FOR MULTI-MAPS ***
	return neighbors
end

local function insertTable(list, row, value)
	local lastRow = #list + 1
	for n = lastRow, row + 1, -1 do
		list[n] = list[n-1]
	end
	list[row] = value
	return list
end


function M.pathfind(map, x1, y1, x2, y2) -- A*

	-- Initialize pathfinding
	local dirs = {{-1,0}, {0,-1}, {1,0}, {0,1}}
	local openList = {{x=x1, y=y1, f=estDistCost(map.avgCost, x1, y1, x2, y2), g=0, h=estDistCost(map.avgCost, x1, y1, x2, y2)}}
	local closedList  = {}
	for x = 1, map.w do
		closedList[x] = {} -- Represent closedList as a 2D grid for easy checking
	end

	-- loop through openList
	repeat
		-- Pull the best cell (last cell in list)
		local n = openList[#openList]
		openList[#openList] = nil
		-- put it on closedList
		closedList[n.x][n.y] = {f=n.f, px=n.px, py=n.py}
		-- Examine the 4 neighbor cells
		local neighbors = getNeighbors(map, n.x, n.y)
		for i = 1, #neighbors do
			local x, y = neighbors[i].x, neighbors[i].y
			if map[x] and map[x][y] then
				-- compute cell's score
				local score = {}
				score.g = n.g + map[x][y].pathCost
				score.h = estDistCost(map.avgCost, x, y, x2, y2)
				score.f = score.g + score.h
				-- check closedList
				if closedList[x][y] and score.f < closedList[x][y].f then closedList[x][y] = nil end
				-- check openList
				local row
				for j = 1, #openList do
					if x == openList[j].x and y == openList[j].y then row = j end
				end
				-- insert new cell into openList
				if not (closedList[x][y] or row) then
					row = 0
					repeat
						row = row + 1
					until row > #openList or score.f > openList[row].f
					openList = insertTable(openList, row, {x=x, y=y, px=n.x, py=n.y, f=score.f, g=score.g, h=score.h})
				end
			end
		end
	until closedList[x2][y2] or #openList == 0

	-- Build path
	local path = nil
	if closedList[x2][y2] then
		path = {cost=0}
		path[1] = {x=x2, y=y2}
		repeat
			local x = path[#path].x
			local y = path[#path].y
			path[#path+1] = {x=closedList[x][y].px, y=closedList[x][y].py}
			path.cost = path.cost + closedList[x][y].f
		until path[#path].x == x1 and path[#path].y == y1
	end
	return path
end

return M
