
-- Quick Links:
-- TLtools (input & serialization):	https://love2d.org/wiki/TLTools
-- Spritebatch info etc:			https://love2d.org/forums/viewtopic.php?f=5&t=82562
-- Maze generation					http://pcg.wikidot.com/pcg-algorithm:maze
-- Tunnel generation				http://www.roguebasin.com/index.php?title=Basic_directional_dungeon_generation
	-- this could be tweaked to "elbow" by applying the algorithm twice (once vertical, once horizontal) on the same terrain grid


---- TO DO ----
-- Maps
	-- AFTER global pathfinding, replace world map Tunnels with actually generated tunnels!  :)
	-- No more "default" tile
		-- Center player when possible
		-- Draw towards edge otherwise
	-- usually don't need to pass "map" as an argument - it's a global variable
	-- generate maps for cities, towns, ruins, caves, dungeons, etc
		-- use location value to determine "magnitude" of the map (size, merchant wealth, equipment level, etc)
	-- Only display discovered tiles
	-- Use a spritebatch?
-- UI
	-- Window Title
		-- get starting seed, list it in window title
		-- also list number of locations (to verify they're all accessible)
	-- reserve part of window for UI like minimap, health, stats, etc
		-- don't let minimap cover world map, if possible
	-- Allow minimap (and rest of UI) to be toggled on/off ?
	-- Add on-screen controls for touch interfaces
-- Minimap
	-- Display only discovered tiles
	-- add blinking cursor to player's location, etc
	-- spritebatch instead of canvas?



-- MAJOR GAME FUNCTIONS --

function getValue(array, x, y, z, default) -- *** Eventually delete this
	return array[x] and array[x][y] and array[x][y][z] or default
end

function changeMap(door)

	-- how to display FEATURES on minimap??

	-- Switch maps
	currentMap = maps[door.map]
	player.x = door.x
	player.y = door.y

	-- Create minimap
	local width, height = love.graphics.getDimensions()
	minimap = {x=width-currentMap.w-10, y=height-currentMap.h-10, sx=1, sy=1}
	minimap.canvas = love.graphics.newCanvas(currentMap.w, currentMap.h)
    love.graphics.setCanvas(minimap.canvas)
        love.graphics.clear()
		for x = 1, currentMap.w do
			for y = 1, currentMap.h do
				love.graphics.setColor(mapColors[currentMap[x][y].color])
				love.graphics.points(x, y)
			end
		end
    love.graphics.setCanvas()

	-- Keep it lean
	collectgarbage() -- is this the slow-down?? ***
end

function newGame()

	-- Initialize game creation
--~ 	love.math.setRandomSeed(10) -- HACK ***

	local width, height = 120, 120
	player = {map=1}

	-- Create world map
	maps = map.createNewWorld(width, height)

	-- Begin play
	changeMap(player)
end

function loadSpriteSheet(width, height)
	local tilesX = (width - 1) / (tileSize + 1)
	local tilesY = (height - 1) / (tileSize + 1)
	local spriteArray = {}
	for x = 1, tilesX do
		for y = 1, tilesY do
			spriteArray[(x-1) * tilesY + y] = love.graphics.newQuad((x-1)*(tileSize+1)+1, (y-1)*(tileSize+1)+1, tileSize, tileSize, width, height)
		end
	end
	return spriteArray
end


-- GUI FUNCTIONS

function playerMove(dx, dy)
	-- Validate requested direction
	local target = {x=player.x + dx, y = player.y + dy}
	if not getValue(currentMap, target.x, target.y, "terrain") then return end
	-- check for Creature at target location
		-- hostile creature = attack
		-- friendly creature = talk
	-- Is movement okay?
	if currentMap[target.x][target.y].hasFloor and not currentMap[target.x][target.y].hasWall then
		if currentMap[target.x][target.y].door then
			local door = currentMap[target.x][target.y].door
			player.x = door.x
			player.y = door.y
			if currentMap ~= maps[door.map] then changeMap(maps[door]) end
		else
			player.x = target.x
			player.y = target.y
		end
	end
	-- *** if movement not okay, search tile or do nothing ?
end

function love.load()

	-- ***** DEBUG *****
--~ 	love.math.setRandomSeed(1521486252, 0) -- a good Tunnels seed
--~ 	love.math.setRandomSeed(1521483625, 0) -- a good Boats seed
--~ 	love.math.setRandomSeed(1521491856, 0) -- has a tunnel trapped by a dungeon thing -- watch for that corner case *********
	-- ***** DEBUG *****
	
	-- Initialize
	ai = require "ai"
	map = require "map"
	diamondsquare = require "diamondsquare"
	love.keyboard.setKeyRepeat(true)
	seedLow, seedHigh = love.math.getRandomSeed()
	seedLow, seedHigh = tostring(seedLow), tostring(seedHigh)
	countLocations = ""

	-- Constants & indices
	terOcean, terRiver, terDesrt, terGrass, terTrees, terSwamp, terRocks, terMount = 1, 2, 3, 4, 5, 6, 7, 8
	ftrHmlt, ftrTown, ftrCity, ftrCave, ftrRuin, ftrDngn, ftrTunl, ftrBoat = 43, 44, 45, 42, 40, 41, 38, 39
	mapColors = {
		{  0,   0, 255}, -- Ocean
		{ 30, 144, 255}, -- River
		{255, 215,   0}, -- Desrt
		{173, 255,  47}, -- Grass
		{ 34, 139,  34}, -- Trees
		{  0, 255, 255}, -- Swamp
		{200, 200, 200}, -- Rocks
		{100, 100, 100}} -- Mount

	-- Initialize graphics data
	local width, height = love.graphics.getDimensions()
	tileSize = 32
	tileRadiusX, tileRadiusY = math.floor((width - tileSize)/(tileSize*2)), math.floor((height - tileSize)/(tileSize*2))
	screenPaddingX, screenPaddingY = math.floor((width - tileSize*(tileRadiusX*2+1))/2), math.floor((height - tileSize*(tileRadiusY*2+1))/2)

	-- HACK *** use the CONF file to make the window an appropriate size
	--      *** also, if the window is really big (ie. 4k monitors), apply a scalar to ALL graphics to maintain visibility

	-- Load map Terrain graphics
	mapTerrainImage = love.graphics.newImage("terrain.png")
	mapTerrain = loadSpriteSheet(mapTerrainImage:getDimensions())

	-- Load map Feature graphis
	mapFeatureImage = love.graphics.newImage("feature.png")
	mapFeature = loadSpriteSheet(mapFeatureImage:getDimensions())

	-- Load sprite graphics
	spriteImages = love.graphics.newImage("sprites.png")
	playerImage = love.graphics.newQuad(love.math.random(0,3)*tileSize, love.math.random(0,2)*tileSize, tileSize, tileSize, 128, 96) -- HACK *** Random sprite

	-- Generate game
	newGame() -- HACK *** Waiting on UI, Save/Load, etc
end

function love.keypressed(key)
	if     key == "left"  then playerMove(-1, 0)
	elseif key == "up"    then playerMove(0, -1)
	elseif key == "right" then playerMove(1, 0)
	elseif key == "down"  then playerMove(0, 1)
	end
end

function love.draw()
	-- Reset graphics (is this still important?)
	love.graphics.setColor(255, 255, 255, 255)

	-- HACK *** Fix:
		-- Define a box centered around the player
		-- If it's out of bounds, move it to in-bounds
			-- For any dimension that's smaller than the screen, center it in that dimension
		-- Iterate through, and draw the box's contents
	-- HACK ***

	-- Draw map tiles
	for x = player.x - tileRadiusX, player.x + tileRadiusX do
		for y = player.y - tileRadiusY, player.y + tileRadiusY do
			-- Draw terrain
			local tileToDraw = getValue(currentMap, x, y, "terrain", currentMap.terDefault)
			love.graphics.draw(mapTerrainImage, mapTerrain[tileToDraw], screenPaddingX + (x-player.x+tileRadiusX)*tileSize, screenPaddingY + (y-player.y+tileRadiusY)*tileSize)
			-- Draw any map features
			tileToDraw = getValue(currentMap, x, y, "feature")
			if tileToDraw then love.graphics.draw(mapFeatureImage, mapFeature[tileToDraw], screenPaddingX + (x-player.x+tileRadiusX)*tileSize, screenPaddingY + (y-player.y+tileRadiusY)*tileSize) end
			-- Draw player
			if player.x == x and player.y == y then love.graphics.draw(spriteImages, playerImage, screenPaddingX + tileRadiusX*tileSize, screenPaddingY + tileRadiusY*tileSize) end
			-- Draw any other sprites
			-- ****
		end
	end
	-- Draw minimap
    love.graphics.setBlendMode("alpha", "premultiplied")
    love.graphics.draw(minimap.canvas, minimap.x, minimap.y, 0, minimap.sx, minimap.sy)
    love.graphics.setBlendMode("alpha")
	-- Draw FPS -- DEBUG INFO ***
	love.graphics.print("Current FPS: "..tostring(love.timer.getFPS()).."  \tSeed: "..seedLow..", "..seedHigh.."  \t#Locations: "..countLocations, 10, 10)
end
